# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import os
from contextlib import contextmanager
import logging
import datetime

from datatypes import ReflectClass, make_dict, classproperty

from .compat import *
from .interface import get_interface
from .exception import ReleaseMessage, AckMessage
from .config import environ


logger = logging.getLogger(__name__)


class Message(object):
    """
    this is the base class for sending and handling a message

    to add a new message to your application, just subclass this class

    By default, all subclasses will go to the same queue and then when the queue
    is consumed the correct child class will be created and consume the message
    with its .target() method.

    If you would like your subclass to use a different queue then just set .name
    property on the class and it qill use a different queue
    """
    connection_name = ""
    """the name of the connection to use to retrieve the interface"""

    fields = None
    """holds the actual message that will be sent, this is a dict of key/values
    that will be sent. The fields can be set using properties of this class

    example --
        m = Message
        m.foo = 1
        m.bar = 2
        print(m.fields) # {"foo": 1, "bar": 2}
    """

    name = "morp-messages"
    """The queue name, see .get_name()"""

    classpath_key = "_message_classpath"
    """The key that will be used to hold the Message's child class's full classpath,
    see .hydrate()"""

    @classproperty
    def interface(cls):
        return get_interface(cls.connection_name)

    def __init__(self, fields=None, **fields_kwargs):
        self.fields = self.make_dict(fields, fields_kwargs)

    def __getattr__(self, key):
        if hasattr(self.__class__, key):
            return super().__getattr__(key)
        else:
            return self.fields[key]

    def __setattr__(self, key, val):
        if hasattr(self.__class__, key):
            super().__setattr__(key, val)
        else:
            self.fields[key] = val

    def __setitem__(self, key, val):
        self.fields[key] = val

    def __getitem__(self, key):
        return self.fields[key]

    def __contains__(self, key):
        return key in self.fields

    def send(self, **kwargs):
        """send the message using the configured interface for this class"""
        queue_off = environ.DISABLED
        name = self.get_name()
        fields = self.to_interface()
        if queue_off:
            logger.warning("DISABLED - Would have sent {} to {}".format(
                fields,
                name,
            ))

        else:
            logger.info("Sending message with '{}' keys to '{}'".format(
                "', '".join(fields.keys()),
                name
            ))
            self.from_interface(self.interface.send(name=name, fields=fields, **kwargs))

    def send_later(self, delay_seconds, **kwargs):
        """Send the message after delay_seconds

        :param delay_seconds: int, up to 900 (15 minutes) per SQS docs (interface specific)
        """
        kwargs["delay_seconds"] = delay_seconds
        return self.send(**kwargs)

    @classmethod
    def get_name(cls):
        """This is what's used as the official queue name, it takes cls.name
        and combines it with MORP_PREFIX environment variable

        :returns: str, the queue name
        """
        name = cls.name
        if env_name := environ.PREFIX:
            name = "{}-{}".format(env_name, name)
        return name

    @classmethod
    @contextmanager
    def recv(cls, block=True, **kwargs):
        """Try and receive a message, this is usually used as a context manager

        Usually you'll want to use .handle(), since that will automatically call
        the message's .target() method, but if you want to do something custom
        with the message then you can use this method by itself

        :Example:
            with Message.recv() as m:
                print(m.fields)

        :param block: bool, if True this will block until it receives a message or
            until timeout has passed
        :param **kwargs:
            * timeout: int, how long to wait before yielding None
        :returns: generator<Message>
        """
        if block:
            m = None
            kwargs.setdefault('timeout', 20) # 20 is the max long polling timeout per Amazon
            while not m:
                with cls.recv_for(**kwargs) as m:
                    if m:
                        yield m

        else:
            kwargs.setdefault('timeout', 1)
            with cls.recv_for(**kwargs) as m:
                if m:
                    yield m


    @classmethod
    @contextmanager
    def recv_for(cls, timeout, **kwargs):
        """try and receive a message, return None if a message is not received
        within timeout

        This is a semi-internal method, you'll notice both .recv() and .handle()
        use this method. It attempts to get a message and will ack or release the
        message depending on how .target() did

        :param timeout: float|int, how many seconds before yielding None
        :returns: generator<Message>
        """
        i = cls.interface
        name = cls.get_name()
        ack_on_recv = kwargs.pop('ack_on_recv', False)
        logger.debug("Waiting to receive on {} for {} seconds".format(name, timeout))
        fields = i.recv(name, timeout=timeout, **kwargs)
        if fields:
            try:
                yield cls.hydrate(fields)

            except ReleaseMessage as e:
                i.release(name, fields, delay_seconds=e.delay_seconds)

            except AckMessage as e:
                i.ack(name, fields)

            except Exception as e:
                if ack_on_recv:
                    i.ack(name, fields)
                else:
                    i.release(name, fields)

                raise

            else:
                i.ack(name, fields)

        else:
            yield None

    @classmethod
    def handle(cls, count=0, **kwargs):
        """wait for messages to come in and handle them by calling the incoming
        message's target() method

        :Example:
            Message.handle(10) # handle 10 messages by consuming them and calling .target()

        :param count: int, if you only want to handle N messages, pass in count
        :param **kwargs: any other params will get passed to underlying recv methods
        """
        max_count = count
        count = 0
        while not max_count or count < max_count:
            with cls.recv(**kwargs) as m:
                m.target()
                count += 1

    @classmethod
    def create(cls, *args, **kwargs):
        """create an instance of cls with the passed in fields and send it off

        Since this passed *args and **kwargs directly to .__init__, you can override
        the .__init__ method and customize it and this method will inherit the
        child class's changes. And the signature of this method should always match
        .__init__

        :param *args: list[Any], passed directly to .__init__
        :param **kwargs: dict[str, Any], passed directly to .__init__
        """
        instance = cls(*args, **kwargs)
        instance.send()
        return instance

    @classmethod
    def unsafe_clear(cls):
        """clear the whole message queue"""
        n = cls.get_name()
        return cls.interface.unsafe_clear(n)

    @classmethod
    def count(cls):
        """how many messages total (approximately) are in the whole message queue"""
        n = cls.get_name()
        return cls.interface.count(n)

    @classmethod
    def make_dict(cls, fields, fields_kwargs):
        """lot's of methods take a dict or kwargs, this combines those"""
        return make_dict(fields, fields_kwargs)

    @classmethod
    def get_class(cls, classpath):
        """wrapper to make it easier to do this in child classes, which seems to
        happen quite frequently"""
        return ReflectClass.get_class(classpath)

    @classmethod
    def hydrate(cls, fields):
        """This is used by the interface to populate an instance with information
        received from the interface

        :param imessage: InterfaceMessage, the message freshly received from the
            interface, see Interface.create_imessage()
        """
        message_class = cls
        if cls is Message:
            # When a generic Message instance is used to consume messages it
            # will use the passed in classpath to create the correct Message child
            message_class = cls.get_class(fields[cls.classpath_key])

        instance = message_class()
        instance.from_interface(fields)

        return instance

    def to_interface(self):
        """When sending a message to the interface this method will be called

        :returns: dict, the fields
        """
        fields = self.fields
        if self.classpath_key not in fields:
            fields[self.classpath_key] = ReflectClass.get_classpath(self)
        return fields

    def from_interface(self, fields):
        """When receiving a messag from the interface this method will be called

        you can see it in action with .hydrate()

        :param fields: dict, the fields received from the interface
        """
        self.fields.update(fields)

    def target(self):
        """This method will be called from handle() and can handle any processing
        of the message, it should be defined in the child classes"""
        raise NotImplementedError()

    def ack(self):
        """Acknowledge this message has been processed"""
        cls.interface.ack(cls.get_name(), self.to_interface())

    def release(self, **kwargs):
        """Release this message back to the interface so another message instance
        can pick it up
        """
        cls.interface.release(cls.get_name(), self.to_interface(), **kwargs)

    def release_later(self, delay_seconds):
        """If you want to release the message and not have it be visible for some
        amount of time

        :param delay_seconds: int, how many seconds before the message can be
            processed again. The max value is interface specific
        """
        return self.release(delay_seconds=delay_seconds)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
from contextlib import contextmanager
import re
import itertools
import base64

import boto3
from botocore.exceptions import ClientError

from ..compat import *
from .base import Interface


class Region(String):
    """Small wrapper that just makes sure the AWS region is valid"""
    def __new__(cls, region_name):
        if not region_name:
            session = boto3.Session()
            region_name = session.region_name
            if not region_name:
                raise ValueError("No region name found")

        return super().__new__(cls, region_name)

    @classmethod
    def names(cls):
        """Return all available regions for SQS"""
        session = boto3.Session()
        return session.get_available_regions("ec2")


class SQS(Interface):
    """wraps amazon's SQS to make it work with our generic interface

    https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs.html
    https://boto.readthedocs.org/en/latest/ref/sqs.html
    http://boto3.readthedocs.io/en/latest/guide/sqs.html
    http://michaelhallsmoore.com/blog/Python-Message-Queues-with-Amazon-Simple-Queue-Service
    http://aws.amazon.com/sqs/
    """
    _connection = None

    def _connect(self, connection_config):
        self.connection_config.options['vtimeout_max'] = 43200 # 12 hours max (from Amazon)

        region = Region(self.connection_config.options.get('region', ''))
        self.connection_config.options['region'] = region

        # https://boto3.amazonaws.com/v1/documentation/api/latest/guide/credentials.html
        # https://github.com/boto/boto3/issues/2360#issuecomment-761526845
        if arn := self.connection_config.options.get("arn", ""):
            sts = boto3.client(
                "sts",
                aws_access_key_id=connection_config.username,
                aws_secret_access_key=connection_config.password
            )

            role_info = sts.assume_role(
                RoleArn=arn,
                RoleSessionName="morp",
            )
            credentials = role_info["Credentials"]

            self._connection = boto3.resource(
                "sqs",
                region_name=region,
                aws_access_key_id=credentials["AccessKeyId"],
                aws_secret_access_key=credentials["SecretAccessKey"],
                aws_session_token=credentials["SessionToken"]
            )

            # we no longer need the sts client so clean it up
            self._close_client(sts)

        else:
            self._connection = boto3.resource(
                'sqs',
                region_name=region,
                aws_access_key_id=connection_config.username,
                aws_secret_access_key=connection_config.password
            )

        self.log("SQS connected to region {}", region)

    def get_connection(self):
        return self._connection

    def _close(self):
        """closes out the client and gets rid of connection"""
        if self._connection:
            client = self._connection.meta.client
            self._close_client(client)

        self._connection = None

    def _close_client(self, client):
        """closes open sessions on client

        this code comes from: 
            https://github.com/boto/botocore/pull/1810

        it closes the connections to fix unclosed connection warnings in py3

        Specifically, this warning:

            ResourceWarning: unclosed <ssl.SSLSocket ...
            ResourceWarning: Enable tracemalloc to get the object allocation tracebac

        see also:
            https://github.com/boto/boto3/issues/454
            https://github.com/boto/botocore/pull/1231

        this might also be a good/alternate solution for this problem:
            https://github.com/boto/boto3/issues/454#issuecomment-335614919

        :param client: an amazon services client whose sessions will be closed
        """
        client._endpoint.http_session._manager.clear()
        for manager in client._endpoint.http_session._proxy_managers.values():
            manager.close()

    def get_attrs(self, **kwargs):
        attrs = {}
        options = self.connection_config.options

        # we use max_timeout here because we will release the message
        # sooner according to our release algo but on exceptional error
        # let's use our global max setting
        vtimeout = options.get('max_timeout')
        if vtimeout:
            # if not string fails with:
            # Invalid type for parameter Attributes.VisibilityTimeout, value: 3600,
            # type: <type 'int'>, valid types: <type 'basestring'>
            attrs["VisibilityTimeout"] = String(min(vtimeout, options["vtimeout_max"]))

        for k, v in itertools.chain(options.items(), kwargs.items()):
            if re.match("^[A-Z][a-zA-Z]+$", k):
                attrs[k] = v

        return attrs

    @contextmanager
    def queue(self, name, connection, **kwargs):
        # http://boto3.readthedocs.io/en/latest/reference/services/sqs.html#SQS.Queue
        try:
            q = None

            try:
                q = connection.get_queue_by_name(QueueName=name)
                yield q

            except ClientError as e:
                if self._is_client_error_match(e, ["AWS.SimpleQueueService.NonExistentQueue"]):
                    if kwargs.get("create_queue", True):
                        attrs = self.get_attrs(**kwargs)
                        q = connection.create_queue(
                            QueueName=name,
                            Attributes=attrs 
                        )

                    yield q

                else:
                    raise

        except Exception as e:
            self.raise_error(e)

        finally:
            if q:
                self._close_client(q.meta.client)

    def fields_to_body(self, fields):
        """This base64 encodes the fields because SQS expects a string, not bytes

        https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/sqs/queue/send_message.html

        :param: dict, the fields to send to the backend
        :returns: str, the body, base64 encoded
        """
        body = super().fields_to_body(fields)
        return String(base64.b64encode(body))

    def _send(self, name, connection, body, **kwargs):
        with self.queue(name, connection) as q:
            delay_seconds = kwargs.get('delay_seconds', 0)
            if delay_seconds > 900:
                # https://docs.aws.amazon.com/AWSSimpleQueueService/latest/APIReference/API_SendMessage.html
                self.log("delay_seconds({}) cannot be greater than 900", delay_seconds, level="warning")
                delay_seconds = 900

            # http://boto3.readthedocs.io/en/latest/reference/services/sqs.html#SQS.Queue.send_message
            receipt = q.send_message(MessageBody=body, DelaySeconds=delay_seconds)
            return receipt["MessageId"], receipt

    def _count(self, name, connection, **kwargs):
        ret = 0
        with self.queue(name, connection) as q:
            ret = int(q.attributes.get('ApproximateNumberOfMessages', 0))
        return ret

    def _clear(self, name, connection, **kwargs):
        with self.queue(name, connection) as q:
            try:
                q.purge()

            except ClientError as e:
                if not self._is_client_error_match(e, ["AWS.SimpleQueueService.PurgeQueueInProgress"]):
                    raise

    def _delete(self, name, connection, **kwargs):
        with self.queue(name, connection, create_queue=False) as q:
            if q:
                q.delete()

    def body_to_fields(self, body):
        """Before sending body to parent's body_to_fields() it will base64 decode it

        :param body: str, the body returned from the backend
        """
        return super().body_to_fields(base64.b64decode(body))

    def recv_to_fields(self, _id, body, raw):
        fields = super().recv_to_fields(_id, body, raw)

        # http://boto3.readthedocs.io/en/latest/reference/services/sqs.html#SQS.Queue.receive_messages
        fields["_count"] = int(raw.attributes.get('ApproximateReceiveCount', 1))
#         created_stamp = int(raw.attributes.get('SentTimestamp', 0.0)) / 1000.0
#         if created_stamp:
#             fields["_created"] = Datetime(created_stamp) 

        return fields

    def _recv(self, name, connection, **kwargs):
        timeout = kwargs.get('timeout', None)
        if timeout is not None:
            # http://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/sqs-long-polling.html
            if timeout < 0 or timeout > 20:
                raise ValueError('timeout must be between 1 and 20')

        vtimeout = kwargs.get('vtimeout', None) # !!! I'm not sure this works
        with self.queue(name, connection) as q:
            _id = body = raw = None
            kwargs = {
                "MaxNumberOfMessages": 1,
                "AttributeNames": ["ApproximateReceiveCount", "SentTimestamp"],
            }

            if timeout:
                kwargs["WaitTimeSeconds"] = timeout
            if vtimeout:
                kwargs["VisibilityTimeout"] = min(
                    vtimeout,
                    self.connection_config.options['vtimeout_max']
                )

            msgs = q.receive_messages(**kwargs)
            if msgs:
                raw = msgs[0]
                body = raw.body
                _id = raw.message_id

            return _id, body, raw

    def _release(self, name, fields, connection, **kwargs):
        with self.queue(name, connection) as q:
            # http://stackoverflow.com/questions/14404007/release-a-message-back-to-sqs
            # http://docs.aws.amazon.com/AWSSimpleQueueService/latest/SQSDeveloperGuide/AboutVT.html
            # When you [change] a message's visibility timeout, the new timeout applies
            # only to that particular receipt of the message. ChangeMessageVisibility
            # does not affect the timeout for the queue or later receipts of the message.
            # If for some reason you don't delete the message and receive it again,
            # its visibility timeout is the original value set for the queue.
            delay_seconds = kwargs.get('delay_seconds', 0)
            q.change_message_visibility_batch(Entries=[{
                "Id": fields["_id"],
                "ReceiptHandle": fields["_raw"].receipt_handle,
                "VisibilityTimeout": delay_seconds
            }])

    def _ack(self, name, fields, connection, **kwargs):
        with self.queue(name, connection) as q:
            q.delete_messages(Entries=[
                {
                    'Id': fields["_id"],
                    'ReceiptHandle': fields["_raw"].receipt_handle,
                }
            ])
            # http://boto3.readthedocs.io/en/latest/reference/services/sqs.html#SQS.Message.delete

    def _is_client_error_match(self, e, codes):
        return e.response["Error"]["Code"] in codes




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import


class Error(Exception):
    """error wrapper"""
    def __init__(self, e, exc_info=None):
        self.e = e
        self.exc_info = exc_info
        super(Error, self).__init__(str(e))


class InterfaceError(Error):
    """specifically for wrapping interface errors"""
    pass


class ReleaseMessage(InterfaceError):
    """Can be raised anytime in a recv() with block or target() method to get the
    message to be released

    :param delay_seconds: int, how long until message is released to be processed
        again, the visibility timeout 
    """
    def __init__(self, delay_seconds=0):
        super(ReleaseMessage, self).__init__("")
        self.delay_seconds = delay_seconds


class AckMessage(InterfaceError):
    """Can be raised anytime in a recv() with block or target() method to get the
    message to be acknowledged"""
    def __init__(self):
        super(AckMessage, self).__init__("")




# -*- coding: utf-8 -*-
from contextlib import contextmanager
import fcntl
import errno
import time
import uuid

from datatypes import (
    Dirpath,
    Filepath,
)

from ..compat import *
from .base import Interface


class Dropfile(Interface):
    """Dropfile interface using local files as messages, great for quick prototyping
    or passing messages from the frontend to the backend on the same machine
    """
    _connection = None

    @contextmanager
    def queue(self, name, connection, **kwargs):
        yield connection.child_dir(name, touch=True)

    def _connect(self, connection_config):
        self._connection = Dirpath(connection_config.path, "morp", "queue", touch=True)

    def get_connection(self):
        return self._connection

    def _close(self):
        pass

    def _send(self, name, connection, body, **kwargs):
        with self.queue(name, connection) as queue:
            now = time.time_ns()
            _id = uuid.uuid4().hex

            if delay_seconds := kwargs.get('delay_seconds', 0):
                now += (delay_seconds * 1000000000)

            message = queue.child_file(f"{now}-{_id}-1.txt")
            message.write_bytes(body)
            return _id, message

    def _count(self, name, connection, **kwargs):
        with self.queue(name, connection) as queue:
            return queue.files().count()

    def recv_to_fields(self, _id, body, raw):
        fields = super().recv_to_fields(_id, body, raw)
        fields["_count"] = raw._count
        fields["_body"] = body
        #fields["_created"] = raw.stat[9]
        return fields

    def _recv(self, name, connection, **kwargs):
        _id = body = raw = None
        timeout = kwargs.get('timeout', None) or 0.0
        count = 0.0

        with self.queue(name, connection) as queue:
            while count <= timeout:
                now = time.time_ns()
                for message in queue.files().sort():
                    parts = message.fileroot.split("-")
                    then = int(parts[0])
                    if now > then:
                        fp = message.open("rb+")
                        try:
                            fcntl.flock(fp, fcntl.LOCK_EX | fcntl.LOCK_NB)

                            body = fp.read()
                            if body:
                                _id = parts[1]
                                message.fp = fp
                                message._count = int(parts[2])
                                raw = message
                                break

                            else:
                                # looks like another process got to this message
                                # first, so try and clean it up
                                self._cleanup(fp, message, truncate=False)

                        except OSError as e:
                            pass

                if body:
                    break

                else:
                    count += 0.1
                    if count < timeout:
                        time.sleep(0.1)

        return _id, body, raw

    def _ack(self, name, connection, fields, **kwargs):
        message = fields["_raw"]
        self._cleanup(message.fp, message)

    def _release(self, name, connection, fields, **kwargs):
        delay_seconds = kwargs.get('delay_seconds', 0)

        _id = fields["_id"]
        message = fields["_raw"]
        body = fields["_body"]
        fp = message.fp

        if delay_seconds:
            now = time.time_ns() + (delay_seconds * 1000000000)

            # let's copy the file body to the future and then delete the old message,
            # sadly, because we've got a lock on the file we can't move or copy it,
            # so we're just going to create a new file
            count = fields["_count"] + 1
            dest = Filepath(message.dirname, f"{now}-{_id}-{count}.txt")
            dest.write_bytes(body)
            self._cleanup(fp, message)

        else:
            # release the message back into the queue
            self._cleanup(fp, message, truncate=False, delete=False)

    def _clear(self, name, connection, **kwargs):
        with self.queue(name, connection) as queue:
            queue.clear()

    def _delete(self, name, connection, **kwargs):
        with self.queue(name, connection) as queue:
            queue.delete()

    def _cleanup(self, fp, message, **kwargs):
        # clear the message so other get() requests will move on (this is the
        # one thing we can do with an exclusive lock to tell other processes
        # we have already looked at the file, I wish we could delete under an 
        # exclusive lock)
        if kwargs.get("truncate", True):
            fp.truncate()

        fcntl.flock(fp, fcntl.LOCK_UN)
        fp.close()

        if kwargs.get("delete", True):
            message.delete()




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import logging
import sys
from contextlib import contextmanager
import base64
import datetime
import json

from datatypes import LogMixin, Datetime
from cryptography.fernet import Fernet
import dsnparse

from ..compat import *
from ..config import DsnConnection
from ..exception import InterfaceError


logger = logging.getLogger(__name__)


class InterfaceABC(LogMixin):
    """This abstract base class containing all the methods that need to be implemented
    in a child interface class.

    Child classes should extend Interface (which extends this class). Interface
    contains the public API for using the interface, these methods are broken out
    from Interface just for convenience of seeing all the methods that must be
    implemented
    """
    def _connect(self, connection_config): raise NotImplementedError()

    def get_connection(self): raise NotImplementedError()

    def _close(self): raise NotImplementedError()

    def _send(self, name, connection, body, **kwargs):
        """similar to self.send() but this takes a body, which is the message
        completely encoded and ready to be sent by the backend

        :returns: tuple, (_id, raw) where _id is the message unique id and raw is
            the returned receipt from the backend
        """
        raise NotImplementedError()

    def _count(self, name, connection, **kwargs):
        """count how many messages are currently in the queue

        :returns: int, the rough count, depending on the backend this might not be exact
        """
        raise NotImplementedError()

    def _recv(self, name, connection, **kwargs):
        """
        :returns: tuple, (_id, body, raw) where body is the body that was originally
            passed into send, raw is the untouched object returned from recv, and
            _id is the unique id of the message
        """
        raise NotImplementedError()

    def _ack(self, name, connection, fields, **kwargs): raise NotImplementedError()

    def _release(self, name, connection, fields, **kwargs): raise NotImplementedError()

    def _clear(self, name, connection, **kwargs): raise NotImplementedError()

    def _delete(self, name, connection, **kwargs): raise NotImplementedError()


class Interface(InterfaceABC):
    """base class for interfaces to messaging"""

    connected = False
    """true if a connection has been established, false otherwise"""

    connection_config = None
    """a config.Connection() instance"""

    def __init__(self, connection_config=None):
        self.connection_config = connection_config

    def connect(self, connection_config=None):
        """connect to the interface

        this will set the raw db connection to self.connection
        """

        if self.connected: return self.connected
        if connection_config: self.connection_config = connection_config

        try:
            self.connected = False
            self._connect(self.connection_config)
            self.connected = True
            self.log(f"Connected to {self.__class__.__name__} interface")

        except Exception as e:
            raise self.raise_error(e)

        return self.connected

    def close(self):
        """
        close an open connection
        """
        if not self.connected: return;

        self._close()
        self.connected = False
        self.log(f"Closed Connection to {self.__class__.__name__} interface")

    @contextmanager
    def connection(self, connection=None, **kwargs):
        try:
            if connection:
                yield connection

            else:
                if not self.connected: self.connect()
                try:
                    connection = self.get_connection()
                    yield connection

                except:
                    raise

        except Exception as e:
            self.raise_error(e)

    def fields_to_body(self, fields):
        """This will prepare the fields passed from Message to Interface.send

        prepare a message to be sent over the backend

        :param fields: dict, all the fields that will be sent to the backend
        :returns: bytes, the fully encoded fields
        """
        serializer = self.connection_config.serializer
        if serializer == "pickle":
            ret = pickle.dumps(fields, pickle.HIGHEST_PROTOCOL)

        elif serializer == "json":
            ret = ByteString(json.dumps(fields))

        key = self.connection_config.key
        if key:
            logger.debug("Encrypting fields")
            f = Fernet(key)
            ret = f.encrypt(ret)

        return ret

    def send_to_fields(self, _id, fields, raw):
        """This creates the value that is returned from .send()

        :param _id: str, the unique id of the message, usually created by the backend
        :param fields: dict, the fields that were passed to .send()
        :param raw: mixed: whatever the backend returned after sending body
        :returns: dict: the fields populated with additional keys. If the key begins
            with an underscore then that usually means it was populated internally
        """
        fields["_id"] = _id
        fields["_send_raw"] = raw
        return fields

    def send(self, name, fields, **kwargs):
        """send an interface message to the message queue

        :param name: str, the queue name
        :param fields: dict, the fields to send to the queue name
        :param **kwargs: anything else, this gets passed to self.connection()
        :returns: dict, see .send_to_fields() for what this returns
        """
        if not fields:
            raise ValueError("No fields to send")

        with self.connection(**kwargs) as connection:
            _id, raw = self._send(
                name=name,
                connection=connection,
                body=self.fields_to_body(fields),
                **kwargs
            )
            self.log(f"Message {_id} sent to {name} -- {fields}")
            return self.send_to_fields(_id, fields, raw)

    def count(self, name, **kwargs):
        """count how many messages are in queue name

        :returns: int, a rough count of the messages in the queue, this is backend
            dependent and might not be completely accurate
        """
        with self.connection(**kwargs) as connection:
            return int(self._count(name, connection=connection))

    def body_to_fields(self, body):
        """This will prepare the body returned from the backend to be passed
        to Message

        This turns a backend body back to the original fields

        :param body: bytes, the body returned from the backend that needs to be converted
            back into a dict
        :returns: dict, the fields of the original message
        """
        key = self.connection_config.key
        if key:
            logger.debug("Decoding encrypted body")
            f = Fernet(key)
            ret = f.decrypt(ByteString(body))

        else:
            ret = body

        serializer = self.connection_config.serializer
        if serializer == "pickle":
            ret = pickle.loads(ret)

        elif serializer == "json":
            ret = json.loads(ret)

        return ret

    def recv_to_fields(self, _id, body, raw):
        """This creates the value that is returned from .recv()

        :param _id: str, the unique id of the message, usually created by the backend
        :param body: bytes, the backend message body
        :param raw: mixed: whatever the backend fetched from its receive method
        :returns: dict: the fields populated with additional keys. If the key begins
            with an underscore then that usually means it was populated internally
        """
        fields = self.body_to_fields(body)
        fields["_id"] = _id
        fields["_raw"] = raw
        fields["_count"] = 1
        return fields

    def recv(self, name, timeout=None, **kwargs):
        """receive a message from queue name

        :param name: str, the queue name
        :param timeout: integer, seconds to try and receive a message before returning None
        :returns: dict, the fields that were sent via .send populated with additional
            keys (additional keys will usually be prefixed with an underscore), it
            will return None if it failed to fetch (ie, timeout or error)
        """
        with self.connection(**kwargs) as connection:
            _id, body, raw = self._recv(
                name,
                connection=connection,
                timeout=timeout,
                **kwargs
            )
            return self.recv_to_fields(_id, body, raw) if body else None

    def ack(self, name, fields, **kwargs):
        """this will acknowledge that the interface message was received successfully

        :param name: str, the queue name
        :param fields: dict, these are the fields returned from .recv that have
            additional fields that the backend will most likely need to ack the message
        """
        with self.connection(**kwargs) as connection:
            self._ack(name, connection=connection, fields=fields)
            self.log("Message {} acked from {}", fields["_id"], name)

    def release(self, name, fields, **kwargs):
        """release the message back into the queue, this is usually for when processing
        the message has failed and so a new attempt to process the message should be made

        :param name: str, the queue name
        :param fields: dict, these are the fields returned from .recv that have
            additional fields that the backend will most likely need to release the message
        """
        with self.connection(**kwargs) as connection:
            delay_seconds = max(kwargs.get('delay_seconds', 0), 0)
            count = fields.get("_count", 0)

            if delay_seconds == 0:
                if count:
                    max_timeout = self.connection_config.options.get("max_timeout")
                    backoff = self.connection_config.options.get("backoff_multiplier")
                    amplifier = self.connection_config.options.get("backoff_amplifier", count)
                    delay_seconds = min(
                        max_timeout,
                        (count * backoff) * amplifier
                    )

            self._release(name, connection=connection, fields=fields, delay_seconds=delay_seconds)
            self.log(
                "Message {} released back to {} count {}, with delay {}s",
                fields["_id"],
                name,
                count,
                delay_seconds
            )

    def unsafe_clear(self, name, **kwargs):
        """cliear the queue name, clearing the queue removes all the messages from
        the queue but doesn't delete the actual queue

        :param name: str, the queue name to clear
        """
        with self.connection(**kwargs) as connection:
            self._clear(name, connection=connection)
            self.log("Messages cleared from {}", name)

    def unsafe_delete(self, name, **kwargs):
        """delete the queue, this removes messages and the queue

        :param name: str, the queue name to delete
        """
        with self.connection(**kwargs) as connection:
            self._delete(name, connection=connection)
            self.log("Queue {} deleted", name)

    def raise_error(self, e):
        """this is just a wrapper to make the passed in exception an InterfaceError"""
        if isinstance(e, InterfaceError) or hasattr(builtins, e.__class__.__name__):
            raise e

        else:
            raise InterfaceError(e) from e




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import logging
import sys
from contextlib import contextmanager
import base64
import datetime

from cryptography.fernet import Fernet
import dsnparse

from ..compat import *
from ..config import DsnConnection
from ..exception import InterfaceError


logger = logging.getLogger(__name__)


interfaces = {}
"""holds all configured interfaces"""


def get_interfaces():
    global interfaces
    if not interfaces:
        configure_environ()
    return interfaces


def get_interface(connection_name=""):
    """get the configured interface that corresponds to connection_name"""
    global interfaces
    if not interfaces:
        configure_environ()
    return interfaces[connection_name]


def set_interface(interface, connection_name=""):
    """bind an .interface.Interface() instance to connection_name"""
    global interfaces
    interfaces[connection_name] = interface


def find_environ(dsn_env_name='MORP_DSN', connection_class=DsnConnection):
    """Returns Connection instances found in the environment

    configure interfaces based on environment variables

    by default, when morp is imported, it will look for MORP_DSN, and MORP_DSN_N (where
    N is 1 through infinity) in the environment, if it finds them, it will assume they
    are dsn urls that morp understands and will configure connections with them. If you
    don't want this behavior (ie, you want to configure morp manually) then just make sure
    you don't have any environment variables with matching names

    The num checks (eg MORP_DSN_1, MORP_DSN_2) go in order, so you can't do MORP_DSN_1, MORP_DSN_3,
    because it will fail on _2 and move on, so make sure your num dsns are in order (eg, 1, 2, 3, ...)

    :param dsn_env_name: str, the environment variable name to find the DSN, this
        will aslo check for values of <dsn_env_name>_N where N is 1 -> N, so you
        can configure multiple DSNs in the environment and this will pick them all
        up
    :param connection_class: Connection, the class that will receive the dsn values
    :returns: generator<connection_class>
    """
    return dsnparse.parse_environs(dsn_env_name, parse_class=connection_class)


def configure_environ():
    """auto hook to configure the environment"""
    for c in find_environ():
        inter = c.interface
        set_interface(c.interface, c.name)


def configure(dsn, connection_class=DsnConnection):
    """
    configure an interface to be used to send messages to a backend

    you use this function to configure an Interface using a dsn, then you can get
    that interface using the get_interface() method

    :param dsn: string, a properly formatted morp dsn, see DsnConnection for how to format the dsn
    """
    #global interfaces
    c = dsnparse.parse(dsn, parse_class=connection_class)
    set_interface(c.interface, c.name)




# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import hashlib
import os
import base64

import dsnparse
from datatypes import ReflectClass
from datatypes import property as cachedproperty
from datatypes.config import Environ


from .compat import *


class Connection(object):
    """The base connection class, you will most likely always use DsnConnection"""
    name = ""
    """str, the name of this connection, handy when you have more than one used interface (eg, nsq)"""

    username = ""
    """str, the username (if needed)"""

    password = ""
    """str, the password (if needed)"""

    hosts = None
    """list, a list of (hostname, port) tuples"""

    path = None
    """str, the path (if applicable)"""

    interface_name = ""
    """str, full Interface class name -- the interface that will connect to the messaging backend"""

    options = None
    """dict, any other interface specific options you need"""

    @property
    def interface_class(self):
        interface_class = ReflectClass.get_class(self.interface_name)
        return interface_class

    @property
    def interface(self):
        interface_class = self.interface_class
        return interface_class(self)

    @property
    def serializer(self):
        serializer = self.options.get("serializer", "pickle")
        if serializer not in set(["pickle", "json"]):
            raise ValueError(f"Unknown serializer {serializer}")
        return serializer

    @cachedproperty(cached="_key")
    def key(self):
        """string -- an encryption key loaded from options['key'],
        it must be 32 bytes long so this makes sure it is"""
        key = self.options.get('key', "")
        if key:
            # Fernet key must be 32 url-safe base64-encoded bytes
            bs = ByteString(ByteString(key).sha256())
            key = base64.b64encode(bs[:32])
        return key

    def __init__(self, **kwargs):
        """
        set all the values by passing them into this constructor, any unrecognized kwargs get put into .options

        :Example:
            c = Connection(
                interface_name="...",
                hosts=[("host", port), ("host2", port2)],
                some_random_thing="foo"
            )

            print c.port # 5000
            print c.options # {"some_random_thing": "foo"}
        """
        self.options = kwargs.pop('options', {})
        self.hosts = []

        for key, val in kwargs.items():
            if hasattr(self, key):
                setattr(self, key, val)
            else:
                self.options[key] = val

        self.options.setdefault('max_timeout', 3600) # 1 hour to process message
        self.options.setdefault('backoff_multiplier', 5) # failure backoff multiplier

    def get_netlocs(self, default_port):
        return ["{}:{}".format(h[0], default_port if h[1] is None else h[1]) for h in self.hosts]

    def get_option(self, key, default_val):
        return getattr(self.options, key, default_val)


class DsnConnection(Connection):
    """
    Create a connection object from a dsn in the form

        InterfaceName://username:password@host:port?opt1=val1&opt2=val2#connection_name

    example -- connect to amazon SQS

        morp.interface.sqs.SQS://AWS_ID:AWS_KEY@
    """
    def __init__(self, dsn):
        self.dsn = dsn
        d = self.parse(dsn)
        super().__init__(**d)

    def parse(self, dsn):
        d = {'options': {}, 'hosts': []}
        parser = dsnparse.parse(dsn)
        p = parser.fields
        p["dsn"] = parser.parser.dsn

        # get the scheme, which is actually our interface_name
        d['interface_name'] = self.normalize_scheme(p["scheme"])

        dsn_hosts = []
        if p["hostname"]:
            d['hosts'].append((p["hostname"], p.get('port', None)))

        d['options'] = p["query_params"] or {}
        d['name'] = p["fragment"]
        d['username'] = p["username"]
        d['password'] = p["password"]
        d["path"] = p["path"]

        return d

    def normalize_scheme(self, v):
        ret = v
        d = {
            "morp.interface.sqs:SQS": set(["sqs"]),
            "morp.interface.dropfile:Dropfile": set(["dropfile"]),
        }

        kv = v.lower()
        for interface_name, vals in d.items():
            if kv in vals:
                ret = interface_name
                break

        return ret


environ = Environ("MORP_") # Load any MORP_* environment variables

environ.setdefault('DISABLED', False, type=bool)
"""If set to 1 then messages won't actually be sent"""

environ.setdefault('PREFIX', '')
"""This prefix will be used to create the queue, see Message.get_name()"""




#!/usr/bin/env python
from setuptools import setup, find_packages
import re
import os
from codecs import open


name = "morp"
kwargs = {"name": name}

def read(path):
    if os.path.isfile(path):
        with open(path, encoding='utf-8') as f:
            return f.read()
    return ""


vpath = os.path.join(name, "__init__.py")
if os.path.isfile(vpath):
    kwargs["packages"] = find_packages(exclude=["tests", "tests.*", "*_test*", "examples"])
else:
    vpath = "{}.py".format(name)
    kwargs["py_modules"] = [name]
kwargs["version"] = re.search(r"^__version__\s*=\s*[\'\"]([^\'\"]+)", read(vpath), flags=re.I | re.M).group(1)

# https://pypi.org/help/#description-content-type
kwargs["long_description"] = read('README.md')
kwargs["long_description_content_type"] = "text/markdown"

kwargs["tests_require"] = []
kwargs["install_requires"] = ['dsnparse', 'boto3', 'cryptography', 'datatypes']


setup(
    description='Send and receive messages without thinking about it',
    keywords="Amazon AWS SQS messages message-passing",
    author='Jay Marcyes',
    author_email='jay@marcyes.com',
    url='http://github.com/Jaymon/{}'.format(name),
    license="MIT",
    classifiers=[ # https://pypi.python.org/pypi?:action=list_classifiers
        'Development Status :: 4 - Beta',
        'Environment :: Web Environment',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Topic :: Database',
        'Topic :: Software Development :: Libraries',
        'Topic :: Utilities',
        'Programming Language :: Python :: 3',
    ],
    entry_points = {
        'console_scripts': [
            '{} = {}.__main__:console'.format(name, name),
        ],
    },
    #test_suite = "{}_test".format(name),
    **kwargs
)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
from __future__ import unicode_literals, division, print_function, absolute_import
import sys
import argparse
import os
import logging
import logging.config

from morp.compat import *
from morp import Message, __version__


logger = logging.getLogger(__name__)


class QuietAction(argparse.Action):

    OPTIONS = "DIWEC"

    DEST = "quiet"

    HELP_QUIET = ''.join([
        'Selectively turn off [D]ebug, [I]nfo, [W]arning, [E]rror, or [C]ritical, ',
        '(--quiet=DI means suppress Debug and Info), ',
        'use - to invert (--quiet=-EW means suppress everything but Error and warning)',
    ])

    HELP_Q_LOWER = ''.join([
        'Turn off [D]ebug (-q), [I]nfo (-qq), [W]arning (-qqq), [E]rror (-qqqq), ',
        'and [C]ritical (-qqqqq)',
    ])

    HELP_Q_UPPER = 'Turn off all logging: [D]ebug, [I]nfo, [W]arning, [E]rror, and [C]ritical'

    def __init__(self, option_strings, dest, nargs=None, **kwargs):
        kwargs["required"] = False
        if "-Q" in option_strings:
            kwargs["default"] = False
            kwargs["const"] = True
            kwargs["nargs"] = 0
        elif "-q" in option_strings:
            kwargs["required"] = False
            kwargs["nargs"] = 0
            kwargs["default"] = None

        super(QuietAction, self).__init__(option_strings, self.DEST, **kwargs)

    def __call__(self, parser, namespace, values, option_string=""):
        #pout.v(parser, namespace, values, option_string)

        if option_string.startswith("--"):
            self.configure_logging(values)

        elif option_string.startswith("-q"):
            count = len(values) + 1
            val = self.OPTIONS[0:count]
            self.configure_logging(val)

        elif option_string.startswith("-Q"):
            self.configure_logging("")

        else:
            raise ValueError()

        setattr(namespace, self.dest, self)

    def call_q(self, count):
        val = self.OPTIONS[0:count]
        self.configure_logging(val)

    def configure_logging(self, val):
        """see --quiet flag help for what this does

        This was ripped from https://github.com/Jaymon/bang/blob/master/bang/__main__.py#L113"""

        if val.startswith("-"):
            # if we had a subtract, then just remove those from being suppressed
            # so -E would only show errors
            levels = val[1:]
        else:
            levels = "".join(set(self.OPTIONS) - set(val.upper()))

        class LevelFilter(object):
            def __init__(self, levels):
                self.levels = levels.upper()
                self.__level = logging.NOTSET
            def filter(self, logRecord):
                return logRecord.levelname[0].upper() in self.levels

        try:
            # https://docs.python.org/2/library/logging.html
            # https://docs.python.org/2/library/logging.config.html#logging-config-dictschema
            # https://docs.python.org/2/howto/logging.html
            # http://stackoverflow.com/questions/8162419/python-logging-specific-level-only
            d = {
                'version': 1,
                'formatters': {
                    'basic': {
                        #'format': '[%(levelname).1s|%(filename)s:%(lineno)s] %(message)s',
                        'format': '[%(levelname).1s] %(message)s',
                    },
                    'message': {
                        'format': '%(message)s'
                    }
                },
                'handlers': {
                    'stdout': {
                        'level': 'NOTSET',
                        'class': 'logging.StreamHandler',
                        'formatter': 'basic',
                        'filters': ['stdout', 'user'],
                        'stream': 'ext://sys.stdout'
                    },
                    'stderr': {
                        'level': 'WARNING',
                        'class': 'logging.StreamHandler',
                        'formatter': 'basic',
                        'filters': ['stderr', 'user'],
                        'stream': 'ext://sys.stderr'
                    },
                },
                'filters': {
                    'stdout': {
                        '()': LevelFilter,
                        'levels': 'DI',
                    },
                    'stderr': {
                        '()': LevelFilter,
                        'levels': 'WEC',
                    },
                    'user': {
                        '()': LevelFilter,
                        'levels': levels,
                    },
                },
                'root': {
                    'level': 'NOTSET',
                    'handlers': ['stdout', 'stderr'],
                },
                'loggers': {
                    'botocore': {
                        'level': 'WARNING',
                    },
                    'boto3': {
                        'level': 'WARNING',
                    },
                },
                'incremental': False,
                'disable_existing_loggers': False,
            }
            logging.config.dictConfig(d)

        except Exception as e:
            raise

        return val


def console():
    parser = argparse.ArgumentParser(description='Consume Morp messages')
    parser.add_argument("--version", "-V", action='version', version="%(prog)s {}".format(__version__))
    parser.add_argument('--count', "-c", dest='count', type=int, default=0, help='How many messages to consume')
    parser.add_argument('--dir', '--directory', '-d', dest='directory', default=os.getcwd(), help='The base directory')
    parser.add_argument('name', metavar='NAME', nargs='?', default=Message.get_name())
    parser.add_argument(
        '--quiet',
        #nargs='?',
        #const='DIWEC',
        #default='',
        action=QuietAction,
        #type=configure_quiet_logging,
        help=QuietAction.HELP_QUIET,
    )
    parser.add_argument(
        "-q",
        action=QuietAction,
        help=QuietAction.HELP_Q_LOWER,
    )
    parser.add_argument(
        "-Q",
        action=QuietAction,
        help=QuietAction.HELP_Q_UPPER,
    )

    args = parser.parse_args()

    # we want to make sure the directory can be imported from since chances are
    # the message classes live in that directory
    basepath = os.path.abspath(os.path.expanduser(String(args.directory)))
    sys.path.append(basepath)

    Message.name = args.name
    logger.info("Handling {} messages".format(Message.get_name()))
    Message.handle(count=args.count)
    sys.exit(0)


if __name__ == "__main__":
    # allow both imports of this module, for entry_points, and also running this module using python -m pyt
    console()




# -*- coding: utf-8 -*-

from .config import (
    DsnConnection,
    Connection,
)
from .interface import (
    get_interface,
    set_interface,
    get_interfaces,
    configure_environ,
    find_environ,
)
from .message import Message
from .exception import (
    InterfaceError,
    ReleaseMessage,
    AckMessage,
)


__version__ = '5.1.0'

